-- Update version number to 1.0.0. First stable release.
-- Update pgtap tests to evaluate more functionality and handle edge cases that had resulted from bug reports in the past.
-- No code changes, however this file is required to update to mimeo 1.0.0 and higher
